/*
SMS Server Tools 3
Copyright (C) Keijo Kasvi
http://smstools3.kekekasvi.com/

Based on SMS Server Tools 2 from Stefan Frings
http://www.meinemullemaus.de/

This program is free software unless you got it under another license directly
from the author. You can redistribute it and/or modify it under the terms of
the GNU General Public License as published by the Free Software Foundation.
Either version 2 of the License, or (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdarg.h>
#include <errno.h>
#include "extras.h"
#include "cfgfile.h"
#include "smsd_cfg.h"
#include "stats.h"
#include "version.h"
#include "blacklist.h"
#include "whitelist.h"
#include "alarm.h"
#include "logging.h"

#define strcpy2(dest, value) copyvalue(dest, sizeof(dest) -1, value, name)

char *tb_sprintf(char* format, ...)
{
  va_list argp;

  va_start(argp, format);
  vsnprintf(tb, sizeof(tb), format, argp);
  va_end(argp);

  return tb;
}

void startuperror(char* format, ...)
{
  va_list argp;
  char text[2048];

  va_start(argp, format);
  vsnprintf(text, sizeof(text), format, argp);
  va_end(argp);

  // Perhaps more safe way than passing a null pointer to realloc:
  if (startup_err_str == NULL)
  {
    startup_err_str = (char *)malloc(strlen(text) +1);
    if (startup_err_str != NULL)
      *startup_err_str = 0;
  }
  else
    startup_err_str = (char *)realloc((void *)startup_err_str, strlen(startup_err_str) + strlen(text) +1);

  if (startup_err_str != NULL)
  {
    strcat(startup_err_str, text);
    startup_err_count++;
  }
}

// Helper function for copying to prevent buffer overflows:
char *copyvalue(char *dest, int maxlen, char *value, char *keyword)
{
  if (strlen(value) > maxlen)
  {
    if (keyword != NULL)
      startuperror("Too long value in config file for \"%s\" (%s)\n", keyword, value);
    else
      startuperror("Too long value in config file (%s)\n", value);
    return NULL;
  }

  strcpy(dest, value);
  return dest;
}

void initcfg()
{
  int i;
  int j;

  autosplit=3;
  receive_before_send=0;
  store_received_pdu=1;
  store_sent_pdu = 1;
  validity_period=255;
  delaytime=10;
  blocktime=60*60;
  blockafter = 3;
  errorsleeptime=10;
  blacklist[0]=0;
  whitelist[0]=0;
  eventhandler[0]=0;
  checkhandler[0]=0;
  alarmhandler[0]=0;
  logfile[0]=0;
  loglevel=-1;  // Will be changed after reading the cfg file if stil -1
  alarmlevel=LOG_WARNING;

  strcpy(d_spool,"/var/spool/sms/outgoing");
  strcpy(d_incoming,"/var/spool/sms/incoming");
  *d_report = 0;
  strcpy(d_checked,"/var/spool/sms/checked");
  d_failed[0]=0;
  d_sent[0]=0;
  d_stats[0]=0;
  stats_interval=60*60;
  stats_no_zeroes=0;
  decode_unicode_text=0;
  internal_combine=0;
  keep_filename = 1;
  store_original_filename = 1;
  regular_run[0] = 0;
  regular_run_interval = 5 * 60;
  admin_to[0] = 0;
  filename_preview = 0;
  incoming_utf8 = 0;
  outgoing_utf8 = 1;
  log_charconv = 0;

  message_count = 0;

  username[0] = 0;
  groupname[0] = 0;

  strcpy(infofile, "/var/run/smsd.working");
  strcpy(pidfile, "/var/run/smsd.pid");

  // 3.0.10: This version should not have terminal initialization here.
  // In >= 3.1 this can be set from the conf file and command line argument is handled different way.
  //terminal = 0;

  os_cygwin = 0;

  for (i=0; i<PROVIDER; i++)
  {
    queues[i].name[0]=0;
    queues[i].directory[0]=0;
    for (j=0; j<NUMS; j++)
      queues[i].numbers[j][0]=0;
  }
  for (i=0; i<DEVICES; i++)
  {
    devices[i].name[0]=0;
    devices[i].device[0]=0;
    devices[i].identity[0] = 0;
    devices[i].identity_header[0] = 0;
    devices[i].incoming=0;
    devices[i].outgoing=1;
    devices[i].pin[0]=0;
    devices[i].pinsleeptime=0;
    devices[i].smsc[0]=0;
    devices[i].baudrate=19200;
    devices[i].send_delay=1;
    devices[i].cs_convert=1;
    devices[i].initstring[0]=0;
    devices[i].initstring2[0]=0;
    devices[i].eventhandler[0]=0;
    devices[i].report=0;
    devices[i].rtscts=1;
    devices[i].read_memory_start=1;
    strcpy(devices[i].mode,"new");
    // Assigned queue name is never a directory name:
    //strcpy(devices[i].queues[0],d_checked);
    //for (j=1; j<PROVIDER; j++)
    for (j=0; j<PROVIDER; j++)
      devices[i].queues[j][0]=0;

    devices[i].primary_memory[0]=0;
    devices[i].secondary_memory[0]=0;
    devices[i].secondary_memory_max=-1;
    devices[i].pdu_from_file[0]=0;
    devices[i].sending_disabled=0;
    devices[i].modem_disabled=0;
    devices[i].decode_unicode_text=-1;
    devices[i].internal_combine=-1;
    devices[i].pre_init = 1;
    devices[i].check_network=1;
    devices[i].admin_to[0]=0;
    devices[i].message_limit=0;
    devices[i].message_count_clear=0;
    devices[i].keep_open = 1; // 0;
    devices[i].rr[0] = 0;
    devices[i].rr_interval = 5 * 60;
    devices[i].rr_cmdfile[0] = 0;
    for (j = 0; j < SIZE_RR_CMD; j++)
      devices[i].rr_cmd[j] = 0;
    devices[i].rr_logfile[0] = 0;
    devices[i].rr_loglevel = LOG_NOTICE;
    devices[i].rr_statfile[0] = 0;
  }
  startup_err_str = NULL;
  startup_err_count = 0;
}

int readcfg()
{
  FILE* File;
  char devices_list[256];
  char name[32];
  char value[PATH_MAX];
  char tmp[PATH_MAX];
  char device_name[32];
  int result;
  int i,j;
  char *p;

  *devices_list = 0;

  File=fopen(configfile,"r");
  if (File)
  {
    /* read global parameter */
    result=my_getline(File,name,sizeof(name),value,sizeof(value));
    while (result==1)
    {
      if (strcasecmp(name,"devices")==0)
        strcpy2(devices_list,value);
      else if (strcasecmp(name,"spool")==0)
        strcpy2(d_spool,value);
      else if (strcasecmp(name,"outgoing")==0)
        strcpy2(d_spool,value);
      else if (strcasecmp(name,"stats")==0)
        strcpy2(d_stats,value);
      else if (strcasecmp(name,"failed")==0)
        strcpy2(d_failed,value);
      else if (strcasecmp(name,"incoming")==0)
        strcpy2(d_incoming,value);
      else if (strcasecmp(name,"checked")==0)
        strcpy2(d_checked,value);
      else if (strcasecmp(name,"sent")==0)
        strcpy2(d_sent,value);
      else if (strcasecmp(name,"mypath")==0)
        // Removed in > 3.0.1 because this is not used. Setting is accepted because of backward compatibility.
        ;
      else if (strcasecmp(name,"delaytime")==0)
        delaytime=atoi(value);
      else if (strcasecmp(name,"blocktime")==0)
        blocktime=atoi(value);
      else if (strcasecmp(name,"blockafter")==0)
        blockafter=atoi(value);
      else if (strcasecmp(name,"stats_interval")==0)
        stats_interval=atoi(value);
      else if (strcasecmp(name,"stats_no_zeroes")==0)
        stats_no_zeroes=yesno(value);
      else if (strcasecmp(name,"errorsleeptime")==0)
        errorsleeptime=atoi(value);
      else if (strcasecmp(name,"eventhandler")==0)
        strcpy2(eventhandler,value);
      else if (strcasecmp(name,"checkhandler")==0)
        strcpy2(checkhandler,value);	
      else if (strcasecmp(name,"alarmhandler")==0)
        strcpy2(alarmhandler,value);
      else if (strcasecmp(name,"blacklist")==0)
        strcpy2(blacklist,value);
      else if (strcasecmp(name,"whitelist")==0)
        strcpy2(whitelist,value);
      else if (strcasecmp(name,"logfile")==0)
        strcpy2(logfile,value);
      else if (strcasecmp(name,"loglevel")==0)
        loglevel=atoi(value);
      else if (strcasecmp(name,"alarmlevel")==0)
        alarmlevel=atoi(value);
      else if (strcasecmp(name,"autosplit")==0)
	autosplit=atoi(value);
      else if (strcasecmp(name,"receive_before_send")==0)
        receive_before_send=yesno(value);
      else if (strcasecmp(name,"store_received_pdu")==0)
        store_received_pdu=atoi(value);
      else if (strcasecmp(name,"store_sent_pdu")==0)
        store_sent_pdu=atoi(value);
      else if (strcasecmp(name,"validity")==0)
      {
        int tmpvalue = 0;

        tmpvalue = parse_validity(value, -1);
        if (tmpvalue == -1)
          startuperror("Invalid validity period in config file: %s\n",value);
        else
          validity_period=tmpvalue;
      }
      else if (strcasecmp(name,"decode_unicode_text")==0)
        decode_unicode_text=yesno(value);
      else if (strcasecmp(name,"internal_combine")==0)
        internal_combine=yesno(value);
      else if (strcasecmp(name,"incoming_utf8")==0)
        incoming_utf8=yesno(value);
      else if (strcasecmp(name,"outgoing_utf8")==0)
        outgoing_utf8=yesno(value);
      else if (strcasecmp(name,"log_charconv")==0)
        log_charconv=yesno(value);
      else if (strcasecmp(name,"user")==0)
        strcpy2(username,value);
      else if (strcasecmp(name,"group")==0)
        strcpy2(groupname,value);
      else if (strcasecmp(name,"infofile")==0)
        strcpy2(infofile,value);
      else if (strcasecmp(name,"pidfile")==0)
        strcpy2(pidfile,value);
      else if (strcasecmp(name,"os_cygwin")==0)
        os_cygwin=yesno(value);
      else
        startuperror("Unknown variable in config file: %s\n",name);
      result=my_getline(File,name,sizeof(name),value,sizeof(value));
    }
    if (result==-1)
      startuperror("Syntax Error in config file: %s\n",value);

    /* read queue-settings */
    if (gotosection(File, "queues") || gotosection(File, "queue"))
    {
/*
      i=0;
      result=my_getline(File,name,sizeof(name),value,sizeof(value));
      while ((result==1) && (i<PROVIDER))
      {
        strcpy2(queues[i].name,name);
        strcpy2(queues[i].directory,value);
        i++;
        result=my_getline(File,name,sizeof(name),value,sizeof(value));
      }
*/
      // 3.1beta3, 3.0.9: inform if there is too many queues defined.
      for (i = 0; ; i++)
      {
        if ((result = my_getline(File, name, sizeof(name), value, sizeof(value))) != 1)
          break;
        if (i >= PROVIDER)
        {
          startuperror("Too many queues defined.\n");
          break;
        }
        strcpy2(queues[i].name, name);
        strcpy2(queues[i].directory, value);
      }

      if (result==-1)
        startuperror("Syntax Error in config file: %s\n",value);
    }

    /* read provider-settings */
    if (gotosection(File, "providers") || gotosection(File, "provider"))
    {
      result=my_getline(File,name,sizeof(name),value,sizeof(value));
      while (result==1)
      {
        i=getqueue(name,tmp);
        if (i>=0)
        {
/*
          for (j=1; j<=NUMS; j++)
	  {
	    if (getsubparam(value,j,tmp,sizeof(tmp)))
	    {
#ifdef DEBUGMSG
  printf("!! queues[%i].numbers[%i]=%s\n",i,j-1,tmp);
#endif
	      strcpy2(queues[i].numbers[j-1],tmp);
	    }
	    else
	      break;
	  }
*/
          // 3.1beta3, 3.0.9: inform if there is too many parameters.
          for (j = 1; ; j++)
          {
            if (getsubparam(value, j, tmp, sizeof(tmp)))
            {
              if (j > NUMS)
              {
                startuperror("Too many parameters for provider %s.\n", name);
                break;
              }
              // 3.1beta4, 3.0.9: remove whitespaces:
              p = tmp;
              while (*p)
              {
                if (is_blank(*p))
                  strcpy(p, p +1);
                else
                  p++;
              }
#ifdef DEBUGMSG
  printf("!! queues[%i].numbers[%i]=%s\n",i,j-1,tmp);
#endif
              strcpy2(queues[i].numbers[j-1],tmp);
            }
            else
              break;
          }

        }
        else
          startuperror("Error in config file: missing queue for %s.\n",name);
        result=my_getline(File,name,sizeof(name),value,sizeof(value));
      }
      if (result==-1)
        startuperror("Syntax Error in config file: %s\n",value);
    }

    // If devices_list is empty, getsubparam still returns 1 while getting the first name.
    // Now this list is checked with it's own error message:
    if (devices_list[0] == 0)
      startuperror("There are no devices specified.\n");
    else
    {
      /* read device-settings */
      for (i=0; i<DEVICES; i++)
      {
        if (getsubparam(devices_list,i+1,device_name,sizeof(device_name)))
        {
          if (!gotosection(File,device_name))
	    startuperror("Could not find device [%s] in config file.\n",device_name);
          else
          {
  	    strcpy2(devices[i].name,device_name);
            result=my_getline(File,name,sizeof(name),value,sizeof(value));
	    while (result==1)
	    {
	      if (strcasecmp(name,"device")==0)
	        strcpy2(devices[i].device,value);
	      else
              if (strcasecmp(name,"queues")==0)
	      {
                // 3.1beta3, 3.0.9: incorrect definition was used previously, now they are different size.
                //for (j=1; j<=NUMS; j++)
/*
                for (j=1; j<=PROVIDER; j++)
                  if (getsubparam(value,j,tmp,sizeof(tmp)))
                    strcpy2(devices[i].queues[j-1],tmp);
                  else
                    break;
*/
// Not yet in 3.0.9:                strcpy(value, ask_value(device_name, name, value));
                // Inform if there is too many queues defined.
                for (j = 1; ; j++)
                {
                  if (getsubparam(value, j, tmp, sizeof(tmp)))
                  {
                    if (j > PROVIDER)
                    {
                      startuperror("Too many queues defined for device %s.\n", device_name);
                      break;
                    }
                    strcpy2(devices[i].queues[j -1], tmp);
                  }
                  else
                    break;
                }

	      }
	      else
              if (strcasecmp(name,"incoming")==0)
	      {
	        if (strcasecmp(value,"high") ==0)
                   devices[i].incoming=2;
	        else
	        {
	          devices[i].incoming=atoi(value);
	          if (devices[i].incoming==0)  // For backward compatibility to older version with boolean value
	            devices[i].incoming=yesno(value);
	        }
              }
	      else if (strcasecmp(name,"cs_convert")==0)
	        devices[i].cs_convert=yesno(value);
	      else if (strcasecmp(name,"pin")==0)
	        strcpy2(devices[i].pin,value);
	      else if (strcasecmp(name,"pinsleeptime")==0)
	        devices[i].pinsleeptime=atoi(value);
	      else if (strcasecmp(name,"mode")==0)
              {
                if (strcasecmp(value,"ascii")==0)
                  fprintf(stderr,"Ascii mode is not supported anymore.\n");
                if ((strcasecmp(value,"old")!=0) && (strcasecmp(value,"new")!=0))
                  startuperror("Invalid mode=%s in config file.\n",value);
                else
                  strcpy2(devices[i].mode,value);
              }
	      else if (strcasecmp(name,"smsc")==0)
	        strcpy2(devices[i].smsc,value);
	      else if (strcasecmp(name,"baudrate")==0)
	        devices[i].baudrate=atoi(value);
              else if (strcasecmp(name,"send_delay")==0)
                devices[i].send_delay=atoi(value);
              else if (strcasecmp(name,"memory_start")==0)
                devices[i].read_memory_start=atoi(value);
	      else if (strcasecmp(name,"init")==0)
              {
                copyvalue(devices[i].initstring, sizeof(devices[i].initstring) -2, value, name);
                strcat(devices[i].initstring,"\r");
              }
              else if (strcasecmp(name,"init1")==0)
              {
                copyvalue(devices[i].initstring, sizeof(devices[i].initstring) -2, value, name);
                strcat(devices[i].initstring,"\r");
              }
              else if (strcasecmp(name,"init2")==0)
              {
                copyvalue(devices[i].initstring2, sizeof(devices[i].initstring2) -2, value, name);
                strcat(devices[i].initstring2,"\r");
              }
	      else if (strcasecmp(name,"eventhandler")==0)
    	        strcpy2(devices[i].eventhandler,value);
	      else if (strcasecmp(name,"report")==0)
    	        devices[i].report=yesno(value);
	      else if (strcasecmp(name,"rtscts")==0)
    	        devices[i].rtscts=yesno(value);
              else if (strcasecmp(name,"primary_memory")==0)
              {
                while ((p = strchr(value,'\"')) != NULL)
                  strcpy(p, p+1);
  	        strcpy2(devices[i].primary_memory,value);
              }
              else if (strcasecmp(name,"secondary_memory")==0)
              {
                while ((p = strchr(value,'\"')) != NULL)
                  strcpy(p, p+1);
  	        strcpy2(devices[i].secondary_memory,value);
              }
              else if (strcasecmp(name,"secondary_memory_max")==0)
                devices[i].secondary_memory_max=atoi(value);
	      else if (strcasecmp(name,"pdu_from_file")==0)
    	        strcpy2(devices[i].pdu_from_file,value);
	      else if (strcasecmp(name,"sending_disabled")==0)
    	        devices[i].sending_disabled=yesno(value);
              else if (strcasecmp(name,"modem_disabled")==0)
                devices[i].modem_disabled=yesno(value);
	      else if (strcasecmp(name,"decode_unicode_text")==0)
    	        devices[i].decode_unicode_text=yesno(value);
	      else if (strcasecmp(name,"internal_combine")==0)
    	        devices[i].internal_combine=yesno(value);
	      else if (strcasecmp(name,"pre_init")==0)
    	        devices[i].pre_init=yesno(value);
	      else
	        startuperror("Unknown variable in config file for modem %s: %s\n",device_name,name);
	      result=my_getline(File,name,sizeof(name),value,sizeof(value));
	    }
            if (result==-1)
              startuperror("Syntax Error in config file: %s\n",value);
            if ((devices[i].report==1) && (devices[i].incoming==0) && devices[i].outgoing)
              fprintf(stderr,"Warning: Cannot receive status reports because receiving is disabled on modem %s\n",device_name);
          }
        }
        else
          break;
      }
    }
    fclose(File);
    set_alarmhandler(alarmhandler,alarmlevel);
    // if loglevel is unset, then set it depending on if we use syslog or a logfile
    if (loglevel==-1)
    {
      if (logfile[0]==0)
        loglevel=LOG_DEBUG;
      else
        loglevel=LOG_WARNING;
    }
  }
  else
  {
    fprintf(stderr,"Cannot open config file for read.\n");
    return 0;
  }
  return 1;
}


int getqueue(char* name, char* directory) // Name can also be a phone number
{
  int i;
  int j;
#ifdef DEBUGMSG
  printf("!! getqueue(name=%s,... )\n",name);
#endif

  // If no queues are defined, then directory is always d_checked
  if (queues[0].name[0]==0)
  {
    strcpy(directory,d_checked);
#ifdef DEBUGMSG
  printf("!! Returns -2, no queues, directory=%s\n",directory);
#endif
    return -2;
  } 
  // Short number is also accepted as a number:
  if (is_number(name) || (*name == 's' && is_number(name +1)))
  {
#ifdef DEBUGMSG
  printf("!! Searching by number\n");
#endif
    i=0;
    while (queues[i].name[0] && (i<PROVIDER))
    {
      j=0;
      while (queues[i].numbers[j][0] && (j<NUMS))
      {
        if (!strncmp(queues[i].numbers[j],name,strlen(queues[i].numbers[j])))
	{
  	  strcpy(directory,queues[i].directory);
#ifdef DEBUGMSG
  printf("!! Returns %i, directory=%s\n",i,directory);
#endif
	  return i;
	}
	j++;
      }
      i++;
    }
  }
  else
  {
#ifdef DEBUGMSG
  printf("!! Searching by name\n");
#endif
    i=0;
    while (queues[i].name[0] && (i<PROVIDER))
    {
      if (!strcmp(name,queues[i].name))
      {
        strcpy(directory,queues[i].directory);
#ifdef DEBUGMSG
  printf("!! Returns %i, directory=%s\n",i,directory);
#endif
        return i;
      }
      i++;
    }
  }
  /* not found */
  directory[0]=0;
#ifdef DEBUGMSG
  printf("!! Returns -1, not found, name=%s, directory=%s\n", name, directory);
#endif
  return -1;
}

int getdevice(char* name)
{
  int i=0;
  while (devices[i].name[0] && (i<DEVICES))
  {
    if (!strcmp(name,devices[i].name))
      return i;
    i++;
  }
  return -1;
}

void help()
{
  printf("smsd spools incoming and outgoing sms.\n\n");
  printf("Usage:\n");
  printf("         smsd [options]\n\n");
  printf("Options:\n");
  printf("         -cx set config file to x\n");
  printf("         -ix set infofile to x\n");
  printf("         -px set pidfile to x\n");
  printf("         -lx set logfile to x\n");
  printf("         -ux set username to x\n");
  printf("         -gx set groupname to x\n");
  printf("         -h  this help\n");
#ifndef NOSTATS
  printf("         -s  display status monitor\n");
#endif
  printf("         -t  run smsd in terminal\n");
  printf("         -V  print copyright and version\n\n");
  printf("All other options are set by the file /etc/smsd.conf.\n\n");
  printf("Output is written to stdout, errors are written to stderr.\n\n");
  exit(0);
}

void parsearguments(int argc,char** argv)
{
  int result;
  strcpy(configfile,"/etc/smsd.conf");
  printstatus=0;
  terminal = 0;
  arg_infofile[0] = 0;
  arg_pidfile[0] = 0;
  arg_username[0] = 0;
  arg_groupname[0] = 0;
  arg_logfile[0] = 0;

  do
  {
    result=getopt(argc,argv,"sthc:Vi:p:l:u:g:");
    switch (result)
    {
      case 'h': help(); 
                break;
      case 'c': copyvalue(configfile, sizeof(configfile) -1, optarg, "configfile commandline argument");
                break;
      case 's':
#ifndef NOSTATS
                printstatus=1;
                break;
#else
                printf("Status monitor is not included in this compilation.\n");
                exit(0);
#endif
      case 't': terminal = 1;
                break;
      case 'V': printf("Version %s, Copyright (c) Keijo Kasvi, %s@%s.%s, http://smstools3.kekekasvi.com\n",
                       smsd_version,"smstools3","kekekasvi","com");
                printf("Based on SMS Server Tools 2 from Stefan Frings, %s@%s.%s, http://www.meinemullemaus.de\n",
                       "smstools","meinemullemaus","de");
                exit(0);
      case 'i': copyvalue(arg_infofile, sizeof(arg_infofile) -1, optarg, "infofile commandline argument");
                break;
      case 'p': copyvalue(arg_pidfile, sizeof(arg_pidfile) -1, optarg, "pidfile commandline argument");
                break;
      case 'l': copyvalue(arg_logfile, sizeof(arg_logfile) -1, optarg, "logfile commandline argument");
                break;
      case 'u': copyvalue(arg_username, sizeof(arg_username) -1, optarg, "username commandline argument");
                break;
      case 'g': copyvalue(arg_groupname, sizeof(arg_groupname) -1, optarg, "groupname commandline argument");
                break;
    }
  }
  while (result>0);
}

int check_directory(char *dir)
{
  int result = 0;
  char fname[PATH_MAX];
  int fd;
  DIR* dirdata;

  if (dir && *dir)
  {
    if ((dirdata = opendir(dir)) == NULL)
      result = 2;
    else
    {
      closedir(dirdata);
      strcpy(fname, dir);
      if (fname[strlen(fname) -1] != '/')
        strcat(fname, "/");
      strcat(fname, "test.XXXXXX");
      if ((fd = mkstemp(fname)) == -1)
        result = 3;
      else
      {
        close(fd);
        unlink(fname);
      }
    }
  }
  else
    result = 1;
  return result;
}

void remove_lockfiles(char *dir)
{
  DIR* dirdata;
  struct dirent* ent;
  struct stat statbuf;
  char tmpname[PATH_MAX];

  if (dir && *dir)
  {
    if ((dirdata = opendir(dir)))
    {
      while ((ent = readdir(dirdata)))
      {
        sprintf(tmpname, "%s%s%s", dir, (dir[strlen(dir) -1] != '/')? "/" : "", ent->d_name);
        stat(tmpname, &statbuf);
        if (S_ISDIR(statbuf.st_mode) == 0)
          if (strcmp(tmpname +strlen(tmpname) -5, ".LOCK") == 0)
            if (unlink(tmpname) != 0)
              startuperror("Cannot unlink file %s: %s\n", tmpname, strerror(errno));  
      }
      closedir(dirdata);
    }
  }
}

void wrlogfile(int *result, char* format, ...)
{
  va_list argp;
  char text[2048];

  va_start(argp, format);
  vsnprintf(text, sizeof(text), format, argp);
  va_end(argp);

  fprintf(stderr, "%s\n", text);
  writelogfile(LOG_CRIT, "smsd", "%s", text);

  if (result != NULL)
    (*result)++;
}

int startup_check(void)
{
  int result = 0; // number of problems
  char *msg_dir = "%s directory %s cannot be opened.";
  char *msg_file = "%s directory %s is not writable.";
  int i;
  int x;
  int y;

  FILE *fp;
  char tmp[PATH_MAX];
  char *p;
  char *p2;
  //struct stat statbuf;
  //int min_permission = 0750;

  if (startup_err_str)
  {
    wrlogfile(NULL, "There was %i error%s while reading the config file.", startup_err_count, (startup_err_count > 1)? "s" : "");  
    p = startup_err_str;
    while (p && *p)
    {
      if ((p2 = strchr(p, '\n')) != NULL)
        *p2 = 0;
      wrlogfile(&result, "- %s", p);
      p = (p2 != NULL)? p2 +1 : NULL;  
    }
    free(startup_err_str);
    startup_err_str = NULL;
    startup_err_count = 0;
  }

  // After this a lockfile errors are collected to startup_err_str. 

  if ((i = check_directory(d_spool)) == 1)
    wrlogfile(&result, "Spool (outgoing) directory definition is missing.");
  else if (i > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Spool", d_spool);
  else if (i == 0)
    remove_lockfiles(d_spool);

  if ((i = check_directory(d_stats)) > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Stats", d_stats);

  if ((i = check_directory(d_failed)) > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Failed", d_failed);
  else if (i == 0)
    remove_lockfiles(d_failed);

  if ((i = check_directory(d_incoming)) == 1)
    wrlogfile(&result, "Incoming directory definition is missing.");
  else if (i > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Incoming", d_incoming);
  else if (i == 0)
    remove_lockfiles(d_incoming);

  if (queues[0].name[0] == 0)
  {
    if ((i = check_directory(d_checked)) == 1)
      wrlogfile(&result, "Checked directory definition is missing.");
    else if (i > 1)
      wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Checked", d_checked);
    else if (i == 0)
      remove_lockfiles(d_checked);
  }

  if ((i = check_directory(d_sent)) > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Sent", d_sent);
  else if (i == 0)
    remove_lockfiles(d_sent);

  if ((i = check_directory(d_report)) > 1)
    wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Report", d_report);
  else if (i == 0)
    remove_lockfiles(d_report);

  x = 0;
  while (queues[x].name[0] && (x < PROVIDER))
  {
    if ((i = check_directory(queues[x].directory)) == 1)
      wrlogfile(&result, "Queue %s directory definition is missing.", queues[x].name);
    else if (i > 1)
      wrlogfile(&result, (i == 2)? msg_dir : msg_file, "Queue", queues[x].directory);
    else if (i == 0)
      remove_lockfiles(queues[x].directory);

    // Should also check that all queue names have a provider setting too:
    if (queues[x].numbers[0][0] == 0)
      wrlogfile(&result, "Queue %s has no provider number(s) defined.", queues[x].name);
    x++;
  }

  if (*eventhandler)
  {
    if ((fp = fopen(eventhandler, "r")) == NULL)
      wrlogfile(&result, "Eventhandler %s cannot be read: %s", eventhandler, strerror(errno));
    else
    {
      fclose(fp);
      //if (stat(eventhandler, &statbuf) >= 0)
      //  if ((statbuf.st_mode & 0755) < min_permission)
      //    wrlogfile(&result, "Eventhandler %s has incorrect permissions.", eventhandler);
      if (!is_executable(eventhandler))
        wrlogfile(&result, "Eventhandler %s is not executable for smsd.", eventhandler);
    }
  }

  if (*checkhandler)
  {
    if ((fp = fopen(checkhandler, "r")) == NULL)
      wrlogfile(&result, "Checkhandler %s cannot be read: %s", checkhandler, strerror(errno));
    else
    {
      fclose(fp);
      if (!is_executable(checkhandler))
        wrlogfile(&result, "Checkhandler %s is not executable for smsd.", checkhandler);
    }
  }

  if (*alarmhandler)
  {
    if ((fp = fopen(alarmhandler, "r")) == NULL)
      wrlogfile(&result, "Alarmhandler %s cannot be read: %s", alarmhandler, strerror(errno));
    else
    {
      fclose(fp);
      if (!is_executable(alarmhandler))
        wrlogfile(&result, "Alarmhandler %s is not executable for smsd.", alarmhandler);
    }
  }

  if (*regular_run)
  {
    if ((fp = fopen(regular_run, "r")) == NULL)
      wrlogfile(&result, "Regular run %s cannot be read: %s", regular_run, strerror(errno));
    else
    {
      fclose(fp);
      if (!is_executable(regular_run))
        wrlogfile(&result, "Regular run %s is not executable for smsd.", regular_run);
    }
  }

  for (x = 0 ; x < DEVICES; x++)
  {
    if (devices[x].name[0])
    {
      if (devices[x].device[0] == 0)
        wrlogfile(&result, "%s has no device specified.", devices[x].name);

      if (queues[0].name[0])
        if (devices[x].queues[0][0] == 0)
          wrlogfile(&result, "Queues are used, but %s has no queue(s) defined.", devices[x].name);

      if (devices[x].eventhandler[0] && strcmp(eventhandler, devices[x].eventhandler) != 0)
      {
        if ((fp = fopen(devices[x].eventhandler, "r")) == NULL)
          wrlogfile(&result, "%s eventhandler %s cannot be read: %s", devices[x].name, devices[x].eventhandler, strerror(errno));
        else
        {
          fclose(fp);
          if (!is_executable(devices[x].eventhandler))
            wrlogfile(&result, "%s eventhandler %s is not executable for smsd.", devices[x].name, devices[x].eventhandler);
        }
      }

      if (devices[x].pdu_from_file[0])
      {
        strcpy(tmp, devices[x].pdu_from_file);
        if ((p = strrchr(tmp, '/')) != NULL)
        {
          *p = 0;
          if ((i = check_directory(tmp)) > 1)
            wrlogfile(&result, (i == 2)? msg_dir : msg_file, "pdu_from_file", tmp);
        }
      }

      if (devices[x].rr[0])
      {
        if ((fp = fopen(devices[x].rr, "r")) == NULL)
          wrlogfile(&result, "%s rr %s cannot be read: %s", devices[x].name, devices[x].rr, strerror(errno));
        else
        {
          fclose(fp);
          if (!is_executable(devices[x].rr))
            wrlogfile(&result, "%s rr %s is not executable for smsd.", devices[x].name, devices[x].rr);
        }
      }

      if (devices[x].rr_cmdfile[0])
      {
        strcpy(tmp, devices[x].rr_cmdfile);
        if ((p = strrchr(tmp, '/')) != NULL)
        {
          *p = 0;
          if ((i = check_directory(tmp)) > 1)
            wrlogfile(&result, (i == 2)? msg_dir : msg_file, "rr_cmdfile", tmp);
        }
      }

      if (devices[x].rr_logfile[0])
      {
        if ((fp = fopen(devices[x].rr_logfile, "a")) == NULL)
          wrlogfile(&result, "%s rr_logfile %s cannot be written: %s", devices[x].name, devices[x].rr_logfile, strerror(errno));
        else
          fclose(fp);
      }

      if (devices[x].rr_statfile[0])
      {
        if ((fp = fopen(devices[x].rr_statfile, "a")) == NULL)
          wrlogfile(&result, "%s rr_statfile %s cannot be written: %s", devices[x].name, devices[x].rr_statfile, strerror(errno));
        else
          fclose(fp);

        // Devices cannot have the same rr_statfile because it's overwritten by each process.
        for (y = 0 ; y < DEVICES; y++)
        {
          if (y == x)
            continue;
          if (devices[y].name[0])
            if (strcmp(devices[y].rr_statfile, devices[x].rr_statfile) == 0)
              wrlogfile(&result, "Devices %s and %s has the same rr_statfile %s.", devices[x].name, devices[y].name, devices[x].rr_statfile);
        }
      }
    }
  }

  if (*whitelist)
  {
    if ((fp = fopen(whitelist, "r")) == NULL)
      wrlogfile(&result, "Whitelist %s cannot be read: %s", whitelist, strerror(errno));
    else
      fclose(fp);
  }

  if (*blacklist)
  {
    if ((fp = fopen(blacklist, "r")) == NULL)
      wrlogfile(&result, "Blacklist %s cannot be read: %s", blacklist, strerror(errno));
    else
      fclose(fp);
  }

  if (*infofile)
  {
    if ((fp = fopen(infofile, "w")) == NULL)
      wrlogfile(&result, "Infofile %s cannot be created: %s", infofile, strerror(errno));
    else
      fclose(fp);
    unlink(infofile);
  }

  if (store_received_pdu < 0 || store_received_pdu > 3)
    wrlogfile(&result, "Invalid value for store_received_pdu.");

  if (store_sent_pdu < 0 || store_sent_pdu > 3)
    wrlogfile(&result, "Invalid value for store_sent_pdu.");

  if (startup_err_str)
  {
    wrlogfile(NULL, "There was %i error%s while removing .LOCK files.", startup_err_count, (startup_err_count > 1)? "s" : "");  
    p = startup_err_str;
    while (p && *p)
    {
      if ((p2 = strchr(p, '\n')) != NULL)
        *p2 = 0;
      wrlogfile(&result, "- %s", p);
      p = (p2 != NULL)? p2 +1 : NULL;  
    }
    free(startup_err_str);
    startup_err_str = NULL;
    startup_err_count = 0;
  }

  if (result > 0)
  {
    wrlogfile(NULL, "There was %i major problem%s found.", result, (result > 1)? "s" : "");
    fprintf(stderr, "Cannot start. See the log file for details.\n");
  }
  else
  {
    // Report some settings:
    char buffer[1024];
    mode_t mode;
    mode_t m;

    mode = umask(0);
    umask(mode);
    m = 0666 & ~mode;
    sprintf(buffer, "File mode creation mask: 0%o (0%o, %c%c%c%c%c%c%c%c%c).", (int)mode, (int)m,
            (m & 0x100)? 'r':'-', (m & 0x80)? 'w':'-', (m & 0x40)? 'x':'-',
            (m & 0x20)? 'r':'-', (m & 0x10)? 'w':'-',  (m & 0x8)? 'x':'-',
            (m & 0x4)? 'r':'-', (m & 0x2)? 'w':'-', (m & 0x1)? 'x':'-');
    writelogfile(LOG_INFO, "smsd", buffer);
#ifdef DEBUGMSG
  printf("!! %s\n", buffer);
#endif

    if (validity_period < 255)
    {
      char tmp[100];

      report_validity(tmp, validity_period);
      sprintf(buffer, "Default validity period is set to %s.", tmp);
      writelogfile(LOG_INFO, "smsd", buffer);
#ifdef DEBUGMSG
  printf("!! %s\n", buffer);
#endif
    }

  }

  return result;
}
